/*
 * Copyright (c) 1999,2000 Arjan van de Ven
 * Copyright (c) 2000      Alessandro Rubini (rubini@linux.it)
 * Copyright (c) 2003      Davide Ciminaghi (ciminaghi@prosa.it)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif

#ifndef MODULE
#  define MODULE
#endif

static int errno;
#ifndef __OPTIMIZE__
#define __OPTIMIZE__
#endif
#define __KERNEL_SYSCALLS__

#include <linux/config.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/file.h>
#include <linux/wait.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <linux/net.h>
#include <linux/proc_fs.h>
#include <linux/highuid.h>
#include <net/sock.h>
#include <linux/byteorder/generic.h>

#include <linux/string.h>

#include <asm/atomic.h>
#include <asm/unistd.h>
#include <asm/uaccess.h>

#include <arpa/tftp.h> /* protocol information */
#include <linux/module.h>
#include "ktftpd.h"

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Rubini-Ciminaghi");

/*
 * Disable setfsuid, not working yet
 */
#undef KTFTFPD_DO_SETFSUID

/*
 * Default settings for the server
 */
#define KTFTPD_DEFAULT_PORT     69
#define KTFTPD_DEFAULT_DIR      "/tftp"
#ifdef KTFTFPD_DO_SETFSUID
#define KTFTPD_MINPERM  S_IRUGO /* readable for user/group/other */
#define KTFTPD_FSUID            65534
#endif

/*
 * The tftpd port can be passed as a parameter
 */
static int ktftpd_port = 69;
MODULE_PARM(ktftpd_port, "i");
MODULE_PARM_DESC(ktftpd_port, "The ktftpd port (default 69)");
/*
 * Similarly we can pass the root directory
 * We reserve 1 byte more for possibly adding a trailing slash.
 */
static char *ktftpd_root = "/tftp/";
MODULE_PARM(ktftpd_root, "s");
MODULE_PARM_DESC(ktftpd_root, "The ktftpd root directory (default /tftp)");

#ifdef KTFTPD_DO_SETFSUID
/*
 * Uid of daemons when accessing files
 */
static uid_t ktftpd_fsuid = KTFTPD_FSUID;
MODULE_PARM(ktftpd_fsuid, "i");
MODULE_PARM_DESC(ktftpd_fsuid, "The numeric ktftpd fsuid");
#endif


/* Debug and demo switches */
/* Define DEBUG to enable debugging instructions */
#define DEBUG 1

/* Define ENABLE_PROC_FS to enable the proc fs entry */
#define ENABLE_PROC_FS 1

/* Ack timeout, millisecs */
#define TFTP_TIMEOUT 2000

/* Number of retries before giving up */
#define TFTP_NRETRY 5

/* Debug macro */
#ifdef DEBUG
#define debug_print(a, args...) do { \
    printk(KERN_INFO "ktftpd: " a, ## args) ; \
} while(0)
#else /* DEBUG */
#define debug_print(a, args...) 
#endif /* DEBUG */

/* segsize and struct tftphdr are defined in <arpa/tftp.h> */
#define PKTSIZE (SEGSIZE+sizeof(struct tftphdr))

unsigned char ktftpd_buffer[PKTSIZE]; /* buffer for the initial packet */

/*
 * Definitions of error strings
 */
#define KTFTPD_ERRSTR_CREAT "Error during creation of socket; terminating\n"
#define KTFTPD_ERRSTR_BIND "Error binding socket\n"
#define KTFTPD_ERRSTR_SENDTO "Error sending udp packet\n"
#define KTFTPD_ERRSTR_RECVFROM "Error receiving udp packet\n"
#define KTFTPD_ERRSTR_TIMEOUT "Timeout waiting for ack\n"
#define KTFTPD_ERRSTR_MISSINGACK "Missing ack. Giving up transfer\n"
#define KTFTPD_ERRSTR_TOOLONG "Absolute path too long\n"
#define KTFTPD_ERRSTR_ENOMEM "Can't allocate buffer\n"
#define KTFTPD_ERRSTR_BADMODE "Transfer mode not supported\n"
#define KTFTPD_ERRSTR_OPEN "Error opening file\n"
#define KTFTPD_ERRSTR_BADCODE "Bad request code!!\n"
#define KTFTPD_ERRSTR_FSUID "Error changing fsuid!!\n"

/* 
 * Counter of all threads (same name as kHTTPd) 
 */
static atomic_t DaemonCount;
static atomic_t RequestCount;

/* 
 * pid of the listening process 
 **/
static pid_t DaemonPid;

/*
 * Wait here if you want to be woken up when the kthread is dead
 */
DECLARE_WAIT_QUEUE_HEAD(kthread_kill_queue);

/* 
 * set at unload time 
 */
int ktftpd_shutdown = 0; 

/* Prototypes */
/*
 * create a socket
 */
static int ktftpd_create_socket(struct socket **);
/*
 * create a socket and bind it 
 */
static int ktftpd_setup_socket(struct socket **, unsigned short int);
/*
 * Check request 
 */
static int ktftpd_check_request(struct tftp_req *);
/* 
 * Check file permissions 
 */
static int ktftpd_open_file(struct tftp_req *);
/*
 * Send file thread 
 */
static int ktftpd_sendfile(void *);
/*
 * The main thread. This must open and bind a socket. And fork children
 */
int ktftpd_main(void *info);
/*
 * Read function for proc file system
 */
#if (defined CONFIG_PROC_FS) && (defined ENABLE_PROC_FS)
static int ktftpd_read_proc(char *, char **, off_t, int, int *, void *);
#endif
/*
 * File read function for NETASCII
 */
int ktftpd_read_netascii(struct tftp_req *, char *);
/*
 * File read function for OCTET
 */
int ktftpd_read_octet(struct tftp_req *, char *);
/*
 * Update error status
 */
int ktftpd_update_errstatus(struct tftp_req *, int );
/*
 * Setup function for proc file system
 */
#if (defined CONFIG_PROC_FS) && (defined ENABLE_PROC_FS)
/* 
 * Setup proc fs entry 
 */
static void ktftpd_setup_proc_fs_entry(void);
/*
 * Kill proc fs entry
 */
static void ktftpd_kill_proc_fs_entry(void);
#else
#define ktftpd_setup_proc_fs_entry()
#define ktftpd_kill_proc_fs_entry()
#endif

/* wait queue, used in waiting for thread termination */
DECLARE_WAIT_QUEUE_HEAD(ktftpd_wait_threads);

/**
 * ktftpd_init(void)
 *
 * Init function: spawn main thread and exit
 */
int ktftpd_init(void)
{
    /* fork the main thread */
    DaemonPid = kernel_thread(ktftpd_main, NULL /* no arg */,
                  0 /* no clone flags: slower but safer */);
    ktftpd_setup_proc_fs_entry();
    /* nothing else */
    return 0;
}

#ifdef KTFTFPD_DO_SETFSUID
/*
 * static long ktftp_setfsuid(uid_t uid)
 *
 * @uid: new fsuid.
 *
 * This was copied from kernel/sys.c
 * Set the fsuid for the process. Return the old fsuid
 */
static long ktftpd_setfsuid(uid_t uid)
{
        int old_fsuid;

	debug_print("ktftpd_setfsuid called: %d\n", uid);

        old_fsuid = current->fsuid;
        if (uid == current->uid || uid == current->euid ||
            uid == current->suid || uid == current->fsuid || 
            capable(CAP_SETUID))
        {
                if (uid != old_fsuid)
                {
			debug_print("ktftpd_setfsuid : 0\n");
			/*
			 * This test was added to avoid kernel panics
			 * because current->mm can be NULL in our case
			 */
			if(current->mm)
				current->mm->dumpable = 0;
			debug_print("ktftpd_setfsuid : 2\n");
                        wmb();
                }
		debug_print("ktftpd_setfsuid : 3\n");
                current->fsuid = uid;
        }

	debug_print("ktftpd_setfsuid : 1\n");

        /* We emulate fsuid by essentially doing a scaled-down version
         * of what we did in setresuid and friends. However, we only
         * operate on the fs-specific bits of the process' effective
         * capabilities 
         *
         * FIXME - is fsuser used for all CAP_FS_MASK capabilities?
         *          if not, we might be a bit too harsh here.
         */
        
        if (!issecure(SECURE_NO_SETUID_FIXUP)) {
                if (old_fsuid == 0 && current->fsuid != 0) {
                        cap_t(current->cap_effective) &= ~CAP_FS_MASK;
                }
                if (old_fsuid != 0 && current->fsuid == 0) {
                        cap_t(current->cap_effective) |=
                                (cap_t(current->cap_permitted) & CAP_FS_MASK);
                }
        }

        return old_fsuid;
}
#endif

/**
 * ktftpd_notify_start(void)
 *
 * Increment Daemon Counter
 */
static inline void ktftpd_notify_start(void)
{
	debug_print("Notify start\n");
	atomic_inc(&DaemonCount);
}

/**
 * ktftpd_notify_stop(void)
 *
 * Decrement Daemon Counter and wake up any process waiting 
 * for Daemon Counter to become 0
 */
static inline void ktftpd_notify_stop(void)
{
	debug_print("Notify stop\n");
	atomic_dec(&DaemonCount);
	wake_up(&ktftpd_wait_threads);
}

/**
 * ktftpd_recvfrom(struct socket *sock, struct sockaddr_in *addr, 
 *                 unsigned char *buf, long timeout)
 *
 * @socket : socket from which message shall be received
 * @addr : pointer to from addr structure
 * @buf: pointer to output buffer
 * @timeout: timeout in msec
 * 
 * This procedure is used as a replacement for recvfrom(). Actually it is
 * is based on the one in kHTTPd which in turn is based on sys_recvfrom.
 * The "poll" code was taken from datagram_poll()
 * The iov is passed by the caller since it hosts the peer's address,
 * and the buffer is passed by the calles because it can't be global
 * (all threads share the same address space)
 * The last argument specifies a timeout. If timeout==0, timeout is infinite
 * Otherwise waits on the socket's wait queue with timeout=timeout
 * Return: >0 => number of chars read, == 0 => timeout, < 0 => error.
 */
static inline int ktftpd_recvfrom(struct socket *sock,
				  struct sockaddr_in *addr,
				  unsigned char *buf,
				  long timeout)
{
	struct msghdr msg;
	struct iovec iov;
	long jiffies_timeout=0;
	int len=0, readable=0;
	mm_segment_t oldfs;
	
	if (sock->sk==NULL) return 0;
	
	msg.msg_flags = 0;
	msg.msg_name = addr;
	msg.msg_namelen  = sizeof(struct sockaddr_in);
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_iov	 = &iov;
	msg.msg_iovlen = 1;
	msg.msg_iov->iov_base = buf;
	msg.msg_iov->iov_len = PKTSIZE;
	
	if(timeout)
		jiffies_timeout = (timeout*HZ)/1000;
	else
		readable=1;
	
	while(jiffies_timeout>0) {
		struct sock *sk = sock->sk;
		if(sk->err || 
		   !skb_queue_empty(&sk->error_queue) || 
		   sk->shutdown == SHUTDOWN_MASK)
			break;
		if (!skb_queue_empty(&sk->receive_queue) || 
		    (sk->shutdown&RCV_SHUTDOWN)) {
			readable=1;
			break;
		}
		jiffies_timeout = interruptible_sleep_on_timeout(
			sk->sleep,
			jiffies_timeout
			);
	};
	if(!jiffies_timeout && !readable)
		/*
		 * Timeout elapsed
		 */
		return 0;
	if(readable) {
		oldfs = get_fs(); set_fs(KERNEL_DS);
		len = sock_recvmsg(sock,&msg,1024,0);
		set_fs(oldfs);
	}
	return readable ? len : -1;
}


/**
 * ktftpd_sendto(struct socket *sock, struct sockaddr_in *to, 
 *                 unsigned char *buf, int len)
 *
 * @socket : socket from which message shall be received
 * @to : pointer to to addr structure
 * @buf: pointer to output buffer
 * @len: message length
 * 
 * This procedure is used as a replacement for sendto(). Largely taken 
 * from sys_sendto() (see /net/socket.c)
 */
static inline int ktftpd_sendto(struct socket *sock,
				struct sockaddr_in *to,
				unsigned char *buf, int len)
{
	struct msghdr msg;
	struct iovec iov;
	mm_segment_t oldfs;
	
	if (sock->sk==NULL) return 0;
	
	msg.msg_flags = 0;
	msg.msg_name = to;
	msg.msg_namelen  = sizeof(struct sockaddr_in);
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_iov	 = &iov;
	msg.msg_iovlen = 1;
	msg.msg_iov->iov_base = buf;
	msg.msg_iov->iov_len = PKTSIZE;
	
	oldfs = get_fs(); set_fs(KERNEL_DS);
	len = sock_sendmsg(sock,&msg,len);
	set_fs(oldfs);
	return len;
}

    
/**
 * ktftpd_create_socket(struct socket *sock)
 * @sock: pointer to relevant socket structure.
 *
 * Create a socket.
 * Largely taken from kHTTPd (sockets.c)
 */
static int ktftpd_create_socket(struct socket **sock)
{
	int error;
	if((error=sock_create(PF_INET,SOCK_DGRAM,IPPROTO_UDP,sock)) < 0) {
		(void)printk(KERN_ERR KTFTPD_ERRSTR_CREAT);
		return error;
	}
	return 0;
}

/**
 * ktftpd_setup_socket(struct socket *sock, unsigned short port)
 * @sock: pointer to relevant socket structure.
 * @port: port we want to bind to
 *
 * Create a socket and bind it 
 * Largely taken from kHTTPd (sockets.c)
 */
static int ktftpd_setup_socket(struct socket **sock, unsigned short port)
{
	struct sockaddr_in sin;
	int error;
	
	if((error=ktftpd_create_socket(sock)) < 0)
		return error;
	
	/* Now bind the socket */    
	sin.sin_family       = AF_INET;
	sin.sin_addr.s_addr  = INADDR_ANY;
	sin.sin_port         = htons(port);
	
	error = (*sock)->ops->bind(*sock,(struct sockaddr*)&sin,sizeof(sin));
	if (error<0) {
		(void)printk(KERN_ERR KTFTPD_ERRSTR_BIND);
		return error;
	}
	return 0;

}

/**
 * ktftpd_setup_path(struct tftp_req *req)
 * @req: pointer to request structure
 *
 * Setup path of file. FIXME: ktftpd_root must have a trailing slash at 
 * the moment.
 */
int ktftpd_setup_path(struct tftp_req *req)
{
	struct tftphdr *msg = (struct tftphdr *)req->buffer;
	if(msg->th_stuff[0]!='/') {
		if(strlen(ktftpd_root)+strlen(msg->th_stuff)+1 > 
			TFTP_MAX_FPATHLEN) {
			printk(KERN_ERR KTFTPD_ERRSTR_TOOLONG);
			return -1;
		}
		strcpy(req->filename,ktftpd_root);
		strcpy(&req->filename[strlen(req->filename)],msg->th_stuff);
	} else
		strcpy(req->filename,msg->th_stuff);
	return 0;
}

/**
 * ktftpd_open_file(struct tftp_req *req)
 * @req: pointer to request structure
 *
 * Check file permissions and open file
 */
static int ktftpd_open_file(struct tftp_req *req)
{
	debug_print("Checking permissions..");
	if(req->opcode == RRQ)
		req->open_mode = O_RDONLY;
	else if(req->opcode == WRQ)
		req->open_mode = O_CREAT;
	req->filp = filp_open(req->filename, req->open_mode, 0);
	if (IS_ERR(req->filp)) {
		printk(KERN_ERR KTFTPD_ERRSTR_OPEN);
		ktftpd_update_errstatus(req,PTR_ERR(req->filp));
		return -1;
	}
	return 0;
}

/**
 * ktftpd_check_request(char *buffer, struct tftp_req *req)
 * @buffer: pointer to message
 * @req: pointer to request structure
 *
 * Check write or read request. Discard any other messages 
 * Also check file permissions
 */
static int ktftpd_check_request(struct tftp_req *req)
{
	char *c;
	struct tftphdr *msg = (struct tftphdr *)req->buffer;
	debug_print("Checking request\n");
	req->opcode = ntohs(msg->th_opcode);
	req->mode = &msg->th_stuff[strlen(msg->th_stuff) + 1];
	/* 
	 * If path is too long we return an EACCESS. FIXME !!
	 */
	if(ktftpd_setup_path(req) < 0) {
		req->errstatus = EACCESS;
		return req->errstatus;
	}
	printk(KERN_INFO "Opcode = %d\n", req->opcode);
	printk(KERN_INFO "Filename = %s\n", req->filename);
	printk(KERN_INFO "Mode = %s\n", req->mode);
	if(req->opcode!=RRQ && req->opcode!=WRQ) {
		printk(KERN_ERR KTFTPD_ERRSTR_BADCODE);
		req->errstatus = EBADOP;
		return req->errstatus;
	}
	/*
	 * Convert mode to capital letters before proceeding
	 */
	for(c=req->mode; *c!=0 && c<req->req_end; c++)
		if(*c>='a' && *c<='z')
			*c-=32;
	/*
	 * Now decide what function is needed to actually read/write data 
	 * from/to the file
	 */
	if(req->opcode==RRQ) {
		if(!strcmp(req->mode,"NETASCII")) 
			req->readfun=ktftpd_read_netascii;
		else if(!strcmp(req->mode,"OCTET"))
			req->readfun=ktftpd_read_octet;
		else {
			req->readfun=NULL;
			req->errstatus=EUNDEF;
			printk(KERN_ERR KTFTPD_ERRSTR_BADMODE);
		}
	}
	/* Request is OK. Now check perms and open file */
	debug_print("OK");
	return  ktftpd_open_file(req);
}

/**
 * ktftpd_child(struct tftp_req *req)
 * @req: pointer to tftp request structure
 *
 * Create new socket for handling request
 * Fork proper new kernel thread (depending on request type)
 */
static int ktftpd_child(struct tftp_req *req)
{    
	/*struct socket *sock;*/
	req->xstat = REQRECV;
	/* Create new socket and bind, use a new local port */
	if(ktftpd_setup_socket(&req->socket, 0) < 0)
		return -1;
	/* NOW WE MUST CREATE A NEW KERNEL THREAD:
	 * ktftpd_sendfile() OR ktftpd_recvfile() 
	 */
	switch(req->opcode) {
	case RRQ:
		return kernel_thread(ktftpd_sendfile, req, 0);
	default:
		return -1;
	}
}

/**
 * ktftpd_senderror(struct tftp_req *req)
 * @req: pointer to tftp request structure
 *
 * Send an error message to the client
 */
static const char *errmsg[] = {
	"Not defined",
	"File not found",
	"Access violation",
	"Disk full or allocation exceeded",
	"Illegal TFTP operation",
	"Unknown transfer ID",
	"File already exists",
	"No such user"
};

#define KTFTPD_ERR_MSG_LEN(r) (4 + strlen(errmsg[r->errstatus]) + 1)

static int ktftpd_senderror(struct tftp_req *req)
{
	struct tftphdr *msg = (struct tftphdr *)req->buffer;
	struct socket *sock = req->socket;
	msg->th_opcode = htons(ERROR);
	msg->th_u.tu_code = htons(req->errstatus);
	memcpy(msg->th_data, errmsg[req->errstatus], 
	       strlen(errmsg[req->errstatus]));
	msg->th_data[strlen(errmsg[req->errstatus])] = 0;
	return ktftpd_sendto(sock, &req->from, req->buffer,
			     KTFTPD_ERR_MSG_LEN(req));
}

/**
 * ktftpd_update_errstatus(struct tftp_req *req, int code)
 * @req: pointer to tftp request structure
 * @code: error code, if negative.
 *
 * Update req->errstatus and req->xstat according to @code
 */
int ktftpd_update_errstatus(struct tftp_req *req, int code)
{
	if(code < 0) {
		switch(code) {
		case -EPERM:
		case -EACCES:
		case -ENOTDIR:
			req->errstatus = EACCESS;
			req->xstat = DONE;
			break;
		case -ENOENT:
			req->errstatus = ENOTFOUND;
			req->xstat = DONE;
			break;
		case -EINTR:
			break;
		default:
			req->errstatus = EACCESS;
			req->xstat = DONE;
			break;
		}
	}
	return code;
}

/**
 * ktftpd_read_octed(struct tftp_req *req, char *buffer)
 * @req: pointer to tftp request structure
 * @buffer: pointer to output buffer
 *
 * Read a SEGSIZE bytes long buffer from a file doing no translation.
 */
int ktftpd_read_octet(struct tftp_req *req, char *buffer)
{
	mm_segment_t oldfs;
	int len;
	oldfs = get_fs(); set_fs(KERNEL_DS);
	len = req->filp->f_op->read(req->filp,buffer,SEGSIZE,&req->offset);
	set_fs(oldfs);
	ktftpd_update_errstatus(req, len);
	return len;
}

/**
 * ktftpd_read_netascii(struct tftp_req *req, char *buffer, loff_t *offset)
 * @req: pointer to tftp request structure
 * @buffer: pointer to output buffer
 *
 * Read a SEGSIZE bytes long buffer from a file doing netascii translation.
 */
int ktftpd_read_netascii(struct tftp_req *req, char *buffer)
{
	mm_segment_t oldfs;
	int nbytes, len = 0;
	char *ptr=buffer;
	
	/*
	 * pending_cr may not be initialized when the first packet is sent
	 * so if req->offset is 0, then req->pending_cr is not considered
	 */
	if(req->pending_cr && req->offset) {
		*ptr++='\n';
		nbytes=1;
	} else
		nbytes=0;
	req->pending_cr=0;
	/*
	 * FIXME: find a way to avoid reading byte-by-byte
	 * Bytes exceeding SEGSIZE could be saved for the next tx packet
	 * Tx buffer should probably be twice as big as it is, to account
	 * for a buffer made of "\n"'s only
	 */
	for( ; nbytes<SEGSIZE; nbytes++) {
		oldfs = get_fs(); set_fs(KERNEL_DS);
		len = req->filp->f_op->read(req->filp,ptr,1,&req->offset);
		set_fs(oldfs);
		if(len < 0)
			ktftpd_update_errstatus(req, len);
		if(len<=0)
			break;
		if(*ptr++=='\n') {
			if(nbytes<SEGSIZE-1) {
				ptr[-1]='\r';
				*ptr++='\n';
				nbytes++;
				continue;
			} else {
				ptr[-1]='\r';
				req->pending_cr=1;
			}
		}
	}
	return len>=0 ? nbytes : len;
}

/**
 * ktftpd_sendfile(void *_req)
 * @_req: pointer to tftp request structure
 *
 * The thread in charge of a single transfer
 */
static int ktftpd_sendfile(void *_req)
{
	struct tftp_req *req=_req;
	struct socket *sock = req->socket;
	struct tftphdr *msg;
	char *buffer;
	int stat, retry, sent_bytes = 0;
	unsigned short sent_block;
	
	/* Start living. Increment Daemon counter */
	atomic_inc(&DaemonCount);
	
	/* And Request counter too */
	atomic_inc(&RequestCount);
	
	/* Take a name */
	snprintf(current->comm, 16,"ktftpd-%08d", atomic_read(&RequestCount)); 

	/* Request is OK and file can be accessed */
	/* This buffer will be used to send packets */
	buffer = kmalloc(PKTSIZE, GFP_KERNEL);
	
	if (!buffer) {
		printk(KERN_ERR KTFTPD_ERRSTR_ENOMEM);
		return -1; /* error */
	}
	for(req->xstat=SENDING, req->offset=0, sent_block=1,retry=TFTP_NRETRY; 
	    !ktftpd_shutdown && req->xstat==SENDING; 
	    sent_block++, retry=5) {
		while(retry) {
			msg = (struct tftphdr *)buffer;
			msg->th_opcode = htons(DATA);
			msg->th_u.tu_block = htons(sent_block);
			sent_bytes = req->readfun(req, msg->th_data);
			if(sent_bytes < 0)
				ktftpd_senderror(req);
			stat = ktftpd_sendto
				(
					sock, 
					&req->from,
					buffer,
					4+sent_bytes
					);
			if(stat < 0) {
				(void)printk(KERN_ERR KTFTPD_ERRSTR_SENDTO);
				req->xstat=ERROR;
				break;
			}
			stat = ktftpd_recvfrom(
				sock, 
				&req->from, 
				req->buffer,
				TFTP_TIMEOUT
				);
			retry--;
			if(!stat) {
				(void)printk(KERN_ERR KTFTPD_ERRSTR_TIMEOUT);
				continue;
			}
			if(stat < 0) {
				(void)printk(KERN_ERR KTFTPD_ERRSTR_RECVFROM);
				req->xstat=ERROR;
				break;
			}
			msg = (struct tftphdr *)req->buffer;
			if(ntohs(msg->th_opcode)==ACK && 
			   ntohs(msg->th_u.tu_block)==sent_block)
				break;
		}
		if(retry<=0) {
			(void)printk(KERN_ERR KTFTPD_ERRSTR_MISSINGACK);
			req->xstat=ERROR;
			break;
		}
		if(sent_bytes < SEGSIZE)
			req->xstat = DONE;
	}
	/* 
	 * OK we're done now, just release everything and decrement 
	 * daemon counter
	 */
	sock_release(sock);
	filp_close(req->filp, current->files);
	kfree(buffer);
	/* Child also frees its request structure */
	kfree(req);
	ktftpd_notify_stop();
	debug_print("Child thread exiting\n");
	return 0;
}

/**
 * ktftpd_main(void *info)
 * @info: dummy pointer
 *
 * The main thread. This must open and bind a socket. And fork children
 * when a valid request comes in.
 */
int ktftpd_main(void *info)
{
	struct socket *sock = NULL;
	struct sockaddr_in sin;
	struct tftp_req *req;
	int i;
	
	debug_print("Main thread started\n");
	
	ktftpd_notify_start(); /* we are running */
	
	sprintf(current->comm,"ktftpd-main"); /* comm is 16 bytes */
	lock_kernel();   /* This seems to be required for exit_mm */
	exit_mm(current);
	unlock_kernel();
	
	/* close open files as well (stdin/out/err) are most likely open */
	for (i=255; i>=0; i--)
		if (current->files->fd[i])
			close(i);
	debug_print("Daemonized: OK\n");
	
	/* Block all signals except SIGKILL, SIGTERM */
	spin_lock_irq(&current->sigmask_lock);
	siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGTERM));
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);

#ifdef KTFTPD_DO_SETFSUID
	/*
	 * Drop privileges for file access 
	 */
	ktftpd_setfsuid(ktftpd_fsuid);
#endif	

	/* Open and bind a listening socket */
	if(ktftpd_setup_socket(&sock, ktftpd_port) < 0)
		goto out;
	
	debug_print("Setup socket: OK\n");
	
	/* Enter main loop */
	while (!signal_pending(current) && !ktftpd_shutdown) {
		int len;
		/* sleep waiting for a connection, use the global buffer */
		len = ktftpd_recvfrom(sock, &sin, ktftpd_buffer,0);
		
		if (len < sizeof(struct tftphdr) || len < 0)
			continue;
		debug_print("Request received\n");
		/*
		 * we sleep waiting for a data packet: when it is there,
		 * we check it and fork a child to handle it
		 * req will be freed by the child !!
		 */
		req = kmalloc(sizeof(struct tftp_req), GFP_KERNEL);
		if(!req) {
		    debug_print("Couldn't allocate request buffer\n");
		    schedule();
		    continue;
		}
		req->buffer = ktftpd_buffer; 
		req->from = sin;
		req->req_end = &ktftpd_buffer[len-1];
		
		if(ktftpd_check_request(req) < 0) {
			req->socket=sock;
			ktftpd_senderror(req);
			kfree(req);
			debug_print("Bad request\n");
			schedule();
			continue;
		}
		debug_print("Request OK, forking child\n");
		/* fork a child */
		ktftpd_child(req);
		/* reap all pending children */
		while  (waitpid(-1,NULL,__WCLONE|WNOHANG) > 0)
			;
		/* and loop over */
	}
 out:
	/* terminated */
	sock_release(sock);
	ktftpd_notify_stop();
	debug_print("main thread terminated\nDaemonCount=%d", 
		    atomic_read(&DaemonCount));
	return -1; /* error */
}

/**
 * ktftpd_read_proc(char *buf, char **start, off_t offset, int count, 
 *                  int *eof, void *data)
 * @buf: pointer to output area
 * @start: pointer to char ptr where the start of information shall be 
 *         written
 * @offset: offset of data within file
 * @count: number of bytes requested
 * @eof: set to 1 if EOF
 * @data: pointer to driver proprietary data (we don't use it now)
 *
 * The proc_fs read function
 */
#if (defined CONFIG_PROC_FS && defined ENABLE_PROC_FS)
static struct proc_dir_entry *pde;

static int ktftpd_read_proc(char *buf, char **start, off_t offset, int count, 
			    int *eof, void *data)
{
	sprintf(buf, "Daemon counter is %d, ktftpd port is %d, \
ktftpd root is %s.\n Warning: ktftpd root must have a trailing slash\n", 
		atomic_read(&DaemonCount),
		ktftpd_port,
		ktftpd_root);
	*eof = 1;
	return(strlen(buf));
}

/**
 * ktftpd_setup_proc_fs_entry()
 *
 * Setup a ktftpd directory and a status entry
 */
static void ktftpd_setup_proc_fs_entry(void)
{
	pde = proc_mkdir("ktftpd", NULL);
	create_proc_read_entry("status", 0, pde, ktftpd_read_proc, NULL);
}

/**
 * ktftpd_kill_proc_fs_entry()
 *
 * Kill  ktftpd directory and  status entry
 */
static void ktftpd_kill_proc_fs_entry(void)
{
	remove_proc_entry("status", pde);
	remove_proc_entry("ktftpd", NULL);
}
#endif /* CONFIG_PROC_FS && ENABLE_PROC_FS*/


/**
 * ktftpd_cleanup(void)
 *
 * The cleanup function
 */
void ktftpd_cleanup(void)
{
	debug_print("Starting cleanup function\n");
	/* Kill proc_fs entry */
	ktftpd_kill_proc_fs_entry();
	/* 
	 * Something went wrong during init. No need to kill 
	 * any daemon
	 */
	if(!atomic_read(&DaemonCount))
		return;
	/* tell all threads to quit */
	ktftpd_shutdown = 1;
	/* kill the one listening (it would take too much time to exit) */
	if(kill_proc(DaemonPid, SIGTERM, 1) < 0) {
		/* 
		 * Something weird probably happened to the main daemon !!!
		 * Don't wait for it to stop
		 */
		atomic_dec(&DaemonCount);
	}
	/* OK, when we get back here, the daemon must have died */
	debug_print("DaemonPid killed\n");
	/* Wait for other threads to terminate (if some is there) */
	wait_event_interruptible(ktftpd_wait_threads, 
				 !atomic_read(&DaemonCount));
	debug_print("DaemonCount is 0!!\n");
	return;
}

/* define init_module() and cleanup_module() */
module_init(ktftpd_init);
module_exit(ktftpd_cleanup);


/*
 * Exported symbols 
 */
#if 0
EXPORT_SYMBOL (ktftpd_main);
EXPORT_SYMBOL (ktftpd_sendfile);
#endif















